"""JARVIS quality of life utilities."""
import hashlib
from typing import Callable, Tuple, Union

from aiohttp import ClientSession

from jarvis_core.filters import url

DEFAULT_BLOCKSIZE = 8 * 1024 * 1024


class Singleton(object):
    REQUIRED = []
    OPTIONAL = {}

    def __new__(cls, *args: list, **kwargs: dict):
        """Create a new singleton."""
        inst = cls.__dict__.get("inst")
        if inst is not None:
            return inst

        inst = object.__new__(cls)
        inst.init(*args, **kwargs)
        inst._validate()
        cls.__inst__ = inst

        return inst

    def _validate(self) -> None:
        for key in self.REQUIRED:
            if not getattr(self, key, None):
                raise ValueError(f"Missing required key: {key}")

    def init(self, **kwargs: dict) -> None:
        """Initialize the object."""
        for key, value in kwargs.items():
            setattr(self, key, value)

        for key, value in self.OPTIONAL.items():
            if not getattr(self, key, None):
                setattr(self, key, value)


async def hash(
    data: str, method: Union[Callable, str] = hashlib.sha256, size: int = DEFAULT_BLOCKSIZE
) -> Tuple[str, int, str]:
    """
    Hash an object.

    Args:
        hash_func: Hash function, default `hashlib.sha256`
        data: URL or text to hash

    Returns:
        Tuple of hash hex, data size, and content type

    Raises:
        Exception: Raise on status
    """
    # Get hashlib method
    if isinstance(method, str):
        method = getattr(hashlib, method, hashlib.sha256)
    method = method()

    # Hash non-URLs
    if not url.match(data):
        method.update(data.encode("UTF8"))
        return method.hexdigest(), len(data), "string"

    # Hash URL contents
    async with ClientSession() as session:
        resp = await session.get(data)
        resp.raise_for_status()
        content_type = resp.content_type
        data_len = resp.content_length

        async for block in resp.content.iter_chunked(n=size):
            method.update(block)

        return method.hexdigest(), data_len, content_type


def convert_bytesize(b: int) -> str:
    """Convert bytes amount to human readable."""
    b = float(b)
    sizes = ["B", "KB", "MB", "GB", "TB", "PB"]
    size = 0
    while b >= 1024 and size < len(sizes) - 1:
        b = b / 1024
        size += 1
    return "{:0.3f} {}".format(b, sizes[size])


def unconvert_bytesize(size: int, ending: str) -> int:
    """Convert human readable to bytes."""
    ending = ending.upper()
    sizes = ["B", "KB", "MB", "GB", "TB", "PB"]
    if ending == "B":
        return size
    # Rounding is only because bytes cannot be partial
    return round(size * (1024 ** sizes.index(ending)))
